#!/usr/bin/env python3

from fabric import Connection
from pathlib import Path
from wireguard_tools import WireguardConfig, WireguardPeer, WireguardKey
from sys import stderr
from io import StringIO
from json import loads
from argparse import ArgumentParser
from dataclasses import dataclass
from dataclass_wizard import fromdict
from ipaddress import IPv4Interface, IPv4Network, IPv6Interface, IPv6Network, AddressValueError, NetmaskValueError
from itertools import combinations

@dataclass
class WovenArgs:
    config: str

@dataclass
class WovenNode:
    listen_address: str
    listen_gateway: str
    interface_name: str
    routed_ipv4_subnets: list[str]
    routed_ipv6_subnets: list[str]

@dataclass
class WovenConfig:
    ptp_ipv4_network: str
    ptp_ipv6_network: str
    min_port: int
    max_port: int
    nodes: dict[str, WovenNode]
    ptp_ipv4_subnet: int = 30
    ptp_ipv6_subnet: int = 64

def generate_wg_configs(config: WovenConfig):
    try:
        ptp_ipv4_network = IPv4Network(config.ptp_ipv4_network)
    except AddressValueError:
        raise ValueError("invalid IPv4 PtP network address")
    except NetmaskValueError:
        raise ValueError("invalid IPv4 PtP network subnet")
    try:
        ptp_ipv6_network = IPv6Network(config.ptp_ipv6_network)
    except AddressValueError:
        raise ValueError("invalid IPv6 PtP network address")
    except NetmaskValueError:
        raise ValueError("invalid IPv6 PtP network subnet")
    
    cs = { id: Connection(node.listen_address, user = "root") for id, node in config.nodes.items() }
    for c in cs.values():
        c.run(f"rm /etc/wireguard/*-loop.conf")
    
    ptp_ipv4_network_iter = ptp_ipv4_network.subnets(new_prefix = config.ptp_ipv4_subnet)
    ptp_ipv6_network_iter = ptp_ipv6_network.subnets(new_prefix = config.ptp_ipv6_subnet)
    port_iter = iter(range(config.min_port, config.max_port))
    
    for (id_a, node_a), (id_b, node_b) in combinations(config.nodes.items(), 2):
        try:
            ptp_ipv4_network = next(ptp_ipv4_network_iter)
        except StopIteration:
            raise ValueError("not enough IPv4 PtP networks to assign")
        try:
            ptp_ipv6_network = next(ptp_ipv6_network_iter)
        except StopIteration:
            raise ValueError("not enough IPv6 PtP networks to assign")
        try:
            port = next(port_iter)
        except StopIteration:
            raise ValueError("not enough ports to assign")
        
        ipv4_iter = ptp_ipv4_network.hosts()
        try:
            ipv4_a = IPv4Interface(f"{next(ipv4_iter)}/{ptp_ipv4_network.prefixlen}")
            ipv4_b = IPv4Interface(f"{next(ipv4_iter)}/{ptp_ipv4_network.prefixlen}")
        except StopIteration:
            raise ValueError("not enough IPv4 addresses in each PtP network")
        
        ipv6_iter = ptp_ipv6_network.hosts()
        try:
            ipv6_a = IPv6Interface(f"{next(ipv6_iter)}/{ptp_ipv6_network.prefixlen}")
            ipv6_b = IPv6Interface(f"{next(ipv6_iter)}/{ptp_ipv6_network.prefixlen}")
        except StopIteration:
            raise ValueError("not enough IPv6 addresses in each PtP network")
        
        key_a = WireguardKey.generate()
        key_a_pub = key_a.public_key()
        
        key_b = WireguardKey.generate()
        key_b_pub = key_b.public_key()
        
        config_a = WireguardConfig(
            addresses = [ipv4_a, ipv6_a],
            listen_port = port,
            private_key = key_a,
            preup=[f"ip ro add {node_b.listen_address}/32 dev {node_a.interface_name} via {node_a.listen_gateway} metric 10 src {node_a.listen_address}"],
            predown=[f"ip ro del {node_b.listen_address}/32 dev {node_a.interface_name} via {node_a.listen_gateway} metric 10 src {node_a.listen_address}"],
            peers = {
                key_b_pub: WireguardPeer(
                    public_key = key_b_pub,
                    allowed_ips = node_b.routed_ipv4_subnets + node_b.routed_ipv6_subnets,
                    endpoint_host = node_b.listen_address,
                    endpoint_port = port,
                    persistent_keepalive = 20
                )
            }
        )
        
        config_b = WireguardConfig(
            addresses = [ipv4_b, ipv6_b],
            listen_port = port,
            private_key = key_b,
            preup=[f"ip ro add {node_a.listen_address}/32 dev {node_b.interface_name} via {node_b.listen_gateway} metric 10 src {node_b.listen_address}"],
            predown=[f"ip ro del {node_a.listen_address}/32 dev {node_b.interface_name} via {node_b.listen_gateway} metric 10 src {node_b.listen_address}"],
            peers = {
                key_a_pub: WireguardPeer(
                    public_key = key_a_pub,
                    allowed_ips = node_a.routed_ipv4_subnets + node_a.routed_ipv6_subnets,
                    endpoint_host = node_a.listen_address,
                    endpoint_port = port,
                    persistent_keepalive = 20
                )
            }
        )
        
        cs[id_a].put(StringIO(config_a.to_wgconfig(wgquick_format = True)), f"/etc/wireguard/{id_a}-{id_b}-loop.conf")
        cs[id_b].put(StringIO(config_b.to_wgconfig(wgquick_format = True)), f"/etc/wireguard/{id_b}-{id_a}-loop.conf")

def main():
    try:
        parser = ArgumentParser("woven")
        parser.add_argument("-c", "--config", default = "config.json", help = "The path to the config file")
        args = parser.parse_args(namespace = WovenArgs)
        config_path = Path(args.config)
        config = fromdict(WovenConfig, loads(config_path.read_bytes()))
        generate_wg_configs(config)
    except ValueError as e:
        print(f"error: {e}", file = stderr)

if __name__ == "__main__":
    main()
